open Util
open Parsec

let (warning, error) = mk_error "Lib: header.ml"

type head = string * value
and value = | One of string
            | Many of head list

let v_many heads = Many heads
let v_one  head  = One  head

let key = many_1s (none_of "{}: \t")

let value_line = many_1s (choice [char '\\' *> choice [ '\n' <$ char 'n'
                                                      ; '\r' <$ char 'r'
                                                      ; '\n' <$ nl
                                                      ;         any ]
                                 ; none_of "{}"])

let rec value_continuation i = ws *> get_char >>= fun j ->
  if j > i
  then (cons <$> value_line <*> value_continuation i) <|> pure []
  else pure []

let one_value i = (fun x xs -> String.concat " " (x :: xs)) <$>
  ws *> value_line <*> value_continuation i

(* The `fun loc -> loc |>` work around strict evaluation. Just ignore them. *)
let rec head = fun loc -> loc |>
    (ws *> get_char                                     >>= fun ident ->
     key <* ws <* (char ':' <?> "':'")                  >>= fun key   ->
     (v_many <$> top_h) <|> (v_one <$> one_value ident) >>= fun value ->
     pure (key, value))
and top_h = fun loc -> loc |> (ws *> char '{' *> heads <* (char '}' <?> "'}'"))
and heads = fun loc -> loc |> many head

(* Exported.  Parser for the whole header *)
let header = top_h <?> "a header, which begins by the '{' character"

(*
  Search functions.
*)
let key_path_is_ok : string -> bool = fun path ->
  List.fold_left
    (fun b c -> b && not (String.contains path c))
    true
    ['\n'; '\r'; '{'; '}'; ' '; '\t']

let key_path : string -> string list = fun key ->
  key |> Stream.of_string
      |> run_parser
          (Some ("key" ^ key))
          (sep_by (many_s (none_of ":")) (char ':') "")
      |> List.filter ((<>) "")

(* Finds the wanted heading.  May issue warning or raise Not_found *)
let rec heading_from_heads : string list -> head list -> string =
  fun path header -> match (path, header) with
    | []       , _      -> "" (* intended default (empty key path) *)
    | key::path, h_list -> heading_from_value
                             path
                             (h_list |> List.find (fst |- ((=) key))
                                     |> snd)
and heading_from_value : string list -> value -> string =
  fun path -> function
    | One  value -> if path <> [] then raise Not_found; value
    | Many []    -> "" (* intended default ("key : {}") *)
    | Many heads ->
      if path = []
      then warning "key path too short.";
      heading_from_heads path heads

(* Exported: gets the header from the list.
   May throw Not_found and issue warnings *)
let value : string -> head list -> string =
  fun path header ->
    if not (key_path_is_ok path)
    then error ("ill formed key path: " ^ path)
    else heading_from_heads (key_path path) header

(* Exported: 3 shortcuts for getting a header *)
let from_file    key = parse_file    header |- value key
let from_channel key = parse_channel header |- value key
let from_conf    key = from_file key "_ussm/settings"
