(**
   A poor man's Parsec, which only works on streams of characters.

   The type psr is a Monad Plus, with <$>, pure, <*>, >>=, fail, <|>,
   which have the obvious meaning.

   There are a few other combinators and primitive parsers.
*)

open Util

let (warning, error) = mk_error "Parsing"

type loc = { line : int            (* starting at 1 *)
           ; char : int            (* starting at 1 *)
           ; file : string option } (* may be stdin *)
type 'a psr = loc -> char Stream.t -> (loc * 'a)

let run_parser : string option -> 'a psr -> char Stream.t -> 'a =
  fun file psr str ->
    try str |> psr {line = 1; char = 1; file = file} |> snd
    with Stream.Error parse_error -> error parse_error

let parse_file : 'a psr -> string -> 'a = fun psr file_name ->
  try file_name |> open_in
                |> Stream.of_channel
                |> run_parser (Some file_name) psr
  with Sys_error err -> error ("System error: " ^ err)

let parse_string : 'a psr -> string -> 'a = fun psr string ->
  string |> Stream.of_string |> run_parser None psr

let parse_channel psr = Stream.of_channel |- run_parser None psr
let parse_string  psr = Stream.of_string  |- run_parser None psr

let string_of_loc : loc -> string = fun loc ->
  (match loc.file with
    | Some file_name -> "in file " ^ file_name ^ ", "
    | None           -> "")
  ^ "at line "     ^ string_of_int loc.line
  ^ ", character " ^ string_of_int loc.char

let (<$>) : ('a -> 'b) -> 'a psr -> 'b psr =
  fun f psr loc str -> let (loc2, res) = psr loc str in
                       (loc2, f res)

let pure : 'a -> 'a psr = fun x loc str -> (loc, x)

let (<*>) : ('a -> 'b) psr -> 'a psr -> 'b psr =
  fun psr_f psr_x loc str -> let (loc2, f) = psr_f loc  str in
                             let (loc3, r) = psr_x loc2 str in
                             (loc3, f r)

let (>>=) : 'a psr -> ('a -> 'b psr) -> 'b psr = fun psr f loc str ->
  let (loc2, r) = psr loc str in f r loc2 str

let fail : 'a psr = fun _ _ -> raise Stream.Failure

let (<|>) : 'a psr -> 'a psr -> 'a psr = fun primary backup loc -> parser
  | [< r = primary loc >] -> r
  | [< r = backup  loc >] -> r

let (<* ) : 'a psr -> 'b psr -> 'a psr = fun x y ->      const <$> x <*> y
let ( *>) : 'a psr -> 'b psr -> 'b psr = fun x y -> swap const <$> x <*> y
let (<$ ) : 'a     -> 'b psr -> 'a psr = fun x p ->          const x <$> p

let choice : 'a psr list -> 'a psr = fun l -> List.fold_right (<|>) l fail

let (<?>) : 'a psr -> string -> 'a psr = fun psr expectation loc str ->
  try psr loc str with Stream.Failure ->
    raise (Stream.Error ("Parse error, " ^ string_of_loc loc
                         ^ ": expected " ^ expectation))

(*  `fun loc -> loc |>` works around strict evaluation. Ignore it *)
let rec many : 'a psr -> 'a list psr = fun psr -> fun loc -> loc |>
    ((cons <$> psr <*> many psr) <|> pure [])

let many_1 item           = cons <$> item <*> many item
let end_by item sep error = many (item <* (sep <?> error))
let sep_by item sep error = cons <$> item <*> many (sep *> (item <?> error))

let many_s  item = string_of_list <$> many   item
let many_1s item = string_of_list <$> many_1 item

let set_loc : (loc -> loc) -> unit psr = fun f loc str -> (f loc, ())
let inc_char = set_loc (fun loc -> {loc with char = loc.char + 1})
let inc_line = set_loc (fun loc -> {loc with line = loc.line + 1; char = 1})

let get_loc  : loc psr = fun loc str -> (loc, loc)
let get_char : int psr = fun loc str -> (loc, loc.char)
let get_line : int psr = fun loc str -> (loc, loc.line)

(* primitive version, not meant for user use *)
let _char_if : (char -> bool) -> char psr = fun pred loc -> parser
  | [< 'c when pred c >] -> (loc, c)
let _char : char -> unit psr = fun c -> () <$ _char_if ((=) c)

let nl : unit psr = (_char '\r' *> (_char '\n' <|> pure ()) <|> _char '\n')
                    *> inc_line

(* full version. Doesn't parse newlines, increments position *)
let char_if : (char -> bool) -> char psr = fun pred ->
  _char_if (fun c -> pred c && c <> '\n' && c <> '\r') <* inc_char
let char : char -> unit psr = function
  | '\r' | '\n' -> raise (Invalid_argument "Parsing.char cannot parse newlines")
  | '\t'-> _char '\t' *> set_loc (fun loc -> {loc with
                                              char = ((loc.char / 8) + 1) * 8})
  | ch  -> _char ch   *> inc_char

let none_of : string -> char psr = fun excluded ->
  char_if (String.contains (excluded ^ "\n\r") |- not)

let some_of : string -> char psr = fun included ->
  if String.contains included '\n' || String.contains included '\r'
  then raise (Invalid_argument "Parsing.some_of cannot parse newlines")
  else char_if (String.contains included)

let any : char psr = none_of ""
let hs  : unit psr = ignore <$> many (some_of " \t")
let ws  : unit psr = ignore <$> sep_by hs nl ""

(* parses one line. The smart version removes endline comments # like this one *)
let line        = many_s any
let smart_line  = many_s (none_of "#") <* many any

(* Parses all lines, remove the empty lines in the process. *)
let lines       = List.filter ((<>) "") <$> sep_by line nl ""
let smart_lines = List.filter ((<>) "") <$> sep_by smart_line nl ""
