(**
   Creates a hierarchical menu for the web site.

   It is language sensitive.  It looks at the default-language heading
   in the _ussm/settings file.
*)

open Util
open Parsec
open Settings

let (warning, error) = mk_error "ussm-menu"

(* directory listings *)
type lang = (string * string) list (* language, title in that language *)
type dir = (string * file) list       (* string : name of the resource *)
and file = Plain of lang | Dir of dir

let map_dir : (string -> lang -> lang) ->
              (string -> dir  -> dir ) ->
              (dir -> dir) =
  fun fp fd dir -> dir /@ fun (file_name, file) ->
                            (file_name, match file with
                              | Dir   d -> Dir   (fd file_name d)
                              | Plain l -> Plain (fp file_name l))

let rec add_to_dir : string -> string -> dir -> dir =
  let name_clash name _ = error ("add_to_dir: name clash with " ^ name) in
  let update dir key new_element lang subdir =
    let maybe f name = if name = key then f else id in
    if List.mem_assoc key dir
    then map_dir (maybe lang) (maybe subdir) dir
    else (key, new_element) :: dir
  in  fun title_path path dir -> match root_path path with
    | ("/", x)
    | (".", x) -> let (name, lang) = name_lang x in
                  let title = title_path |> parse_file (many_s any) in
                  update dir name (Plain [(lang, title)])
                    (cons (lang, title)) (name_clash name)
    | (x,  xs) -> update dir x (Dir (add_to_dir title_path xs []))
                    (name_clash x) (add_to_dir title_path xs)

let dir_of_file_list : string list -> dir =
  List.fold_left (fun dir path ->
    let title_path = "_ussm/tmp/"
                   ^ (String.sub path 0 (String.rindex path '.'))
                   ^ ".title" in
    add_to_dir title_path path dir)
    []

let rec _sort_dir : string -> dir -> dir =
  let default_sorter (name1, file1) (name2, file2) =
    match file1, file2 with
      | Dir _, Plain _ -> -1 (* directory first *)
      | Plain _, Dir _ ->  1 (* directory first *)
      | _              -> String.compare name1 name2 (* lexical order *) in
  let sort_with_list sort list =
    List.fold_right (fun x  l -> (List.find (fst |- (=) x) list) :: l)
                    ("index"::sort) []
  in
  fun path dir ->
    dir /@ (fun (file_name, file) -> match file with
              | Plain l -> (file_name, Plain l)
              | Dir   d -> (file_name, Dir (_sort_dir (path ^"/"^ file_name) d)))
        |> let menu_file = (path ^ "/_menu") in
           if Sys.file_exists menu_file
           then sort_with_list (parse_file smart_lines menu_file)
           else List.sort default_sorter

let sort_dir : dir -> dir = _sort_dir "."

(* filters the dir by the directory path *)
let path_sees : string -> dir -> dir =
  let split = parse_string (sep_by (many_s (none_of "/")) (char '/') "")
           |- List.filter (fun d -> d <> "." && d <> "") in
  let rec aux = function
    | []           -> map_dir (fun _ -> id) (fun _ d -> d // (fst |- (=) "index"))
    | root :: path -> map_dir (fun _ -> id) (fun name d ->
                                               if name = root
                                               then aux path d
                                               else d // (fst |- (=) "index"))
  in split |- aux

let rec remove_index : dir -> dir = fun dir ->
  dir // (fst |- (<>) "index") |> map_dir (fun _ -> id) (fun _ -> remove_index)

let rec get_lang : file -> lang = function
  | Plain l -> l
  | Dir   d -> try get_lang (List.assoc "index" d)
               with Not_found -> error "index file missing"

let get_title : string -> file -> string = fun lang file ->
  let l = get_lang file in
  try l |> List.assoc lang                  with Not_found ->
  try l |> List.assoc (default_language ()) with Not_found ->
      l |> List.hd |> snd (* this one shall never fail *)

let rec xml_of_dir lang prefix dir =
  dir // (fst |- (<>) "index") /@ (fun (f_name, file) ->
    let dir_mark = match file with Dir _ -> "/" | _ -> "" in
    let sub_path = prefix ^ f_name ^ dir_mark in
    let title    = get_title lang file ^ dir_mark in
    Xml.simple_tag "li"
      [ Xml.Tag ("a", [("href", sub_path)], [Xml.Raw title])
      ; xml_of_file lang sub_path file])
and xml_of_file lang prefix = function
  | Plain _
  | Dir []
  | Dir [("index", _)] -> Xml.Raw ""
  | Dir d              -> Xml.simple_tag "ul" (xml_of_dir lang prefix d)

let main () = match Array.length Sys.argv with
  | 1 ->
    let lst_files = [<>] |> open_process "ussm-source-files"
                         |> run_parser None lines                    in
    let dirs  = lst_files /@ (dir_file  |- fst) |> remove_duplicates in
    let langs = lst_files /@ (name_lang |- snd) |> remove_duplicates in
    let files = lst_files |> dir_of_file_list   |> sort_dir          in

    (* Create the */index.*.menu *)
    dirs /@ (fun path -> (let index = path ^ "/index." in
                          langs /@ (fun lang ->
                            ("_ussm/tmp/" ^ index ^ lang ^ ".menu",
                             files |> path_sees path
                                   |> (fun d -> Dir d)
                                   |> xml_of_file lang "/"
                                   |> Xml.text_of_xml))))
         |> List.concat
         |> List.iter (fun (m_path, m_txt) -> let out = (open_out m_path) in
                                              output_string out m_txt;
                                              close_out out);

    (* (Hard) link the other */menus files to the original */index.*.menu ones *)
    lst_files |> List.iter (fun file ->
                   let dir  = "_ussm/tmp/" ^ (file |> dir_file |> fst) in
                   let lang = file |> name_lang |> snd                 in
                   let inp  = dir ^ "/index." ^ lang ^ ".menu"         in
                   let out  = "_ussm/tmp/" ^remove_ext file ^ ".menu"  in
                   if inp <> out
                   then (if Sys.file_exists out then Unix.unlink out;
                         Unix.link inp out))

  | _ -> error "doesn't take any argument."

let _ = standard_exit main
