(** Utility module *)

let (|-) f g x = g (f x) (* composition *)
let (|>) x f   = f x     (* application *)
let id    x   = x        (* indentity function *)
let const x _ = x        (* constant function generator *)
let swap f x y = f y x   (* swaps the first 2 arguments *)

let (//) l f = List.filter f l
let (/@) l f = List.map    f l
let cons x l = x :: l

(* Shortcut for warning and error functions.
   Usage: let (warning, error) = mk_error "program-name" *)
let mk_error : string ->(  (string -> unit) * (string -> 'a))
  = fun program_name ->
    ((fun  warning ->
        prerr_endline (program_name ^ ": Warning: " ^ warning)),
     (fun error ->
        prerr_endline (program_name ^ ": Error: " ^ error);
        raise Exit))

let (warning, error) = mk_error "Core Lib (util.ml)"

let standard_exit : (unit -> unit) -> unit =
  fun main -> try main () with Exit -> exit 1

(* Shortcut for Str.string_match *)
let string_match : string -> string -> bool = fun regexp string ->
  Str.string_match (Str.regexp regexp) string 0

(* The file name must not be empty, or else. *)
let ((dir_file  : string -> string * string),  (* returns (dirname , filename) *)
     (root_path : string -> string * string)) =(* returns (root_dir, sub_path) *)
  let rec aux regexp path =
    if path = "" || path = "." ||
      '/' = path.[String.length path - 1]  then error ("no file name in " ^ path)
    else if not (String.contains path '/') then (".", path)
    else (ignore (string_match regexp path);
          ((match Str.matched_group 1 path with "" -> "/" | d -> d),
           Str.matched_group 2 path))
  in (aux "^\\(.*\\)/\\([^/]*\\)$",
      aux "^\\([^/]*\\)/\\(.*\\)$")

(* Returns the inner-most directory that contains _ussm/ *)
let rec ussm_root : string -> string option = fun dir ->
  if      Sys.file_exists (dir ^ "/_ussm/") then Some dir
  else if dir = "" || dir = "/"             then None
  else if dir = "."                         then ussm_root (Sys.getcwd ())
  else ussm_root (dir |> dir_file |> fst)

(* removes the extension, if any. (but not the directory path *)
let rec remove_ext : string -> string = fun path ->
  let last = String.length path - 1 in
  if not (String.contains path '.') then path
  else if '.' = path.[last]         then String.sub path 0 last
  else (ignore (string_match "^\\(.*\\)\\..[^.]*$" path);
        Str.matched_group 1 path)

(* creates a processs with command,
   hooks stream_input to its standard input,
   hooks its standard output to the returned stream *)
let open_process : string -> char Stream.t -> char Stream.t =
  fun command stream_input ->
    let (output, input) = Unix.open_process command in
    Stream.iter (output_char input) stream_input;
    Stream.of_channel output

(* overrides Pervasives *)
let open_in : string -> in_channel = fun file_name ->
  try  open_in file_name
  with Sys_error err -> error ("System error: " ^ err)
let open_out : string -> out_channel = fun file_name ->
  try open_out file_name
  with Sys_error err -> error ("System error: " ^ err)

let command = Sys.command |- ignore

let string_of_list : char list -> string =
  fun l ->
    let result = String.create (List.length l) in
    ignore (List.fold_left (fun i x -> result.[i] <- x; i+1) 0 l);
    result

(* Last element of a list *)
let rec last : 'a list -> 'a = function
  | []     -> error "last: empty list"
  | [x]    -> x
  | _ :: l -> last l

(* Every element /but/ the last *)
let rec init : 'a list -> 'a list = function
  | []     -> error "init: empty list"
  | [x]    -> []
  | x :: l -> x :: init l

(* Sorts the list in some order, and remove any duplicate *)
let remove_duplicates : 'a list -> 'a list =
  let rec remove_dups = function
    | []                     -> []
    | [x]                    -> [x]
    | x :: y :: l when x = y -> remove_dups (y :: l)
    | x :: l                 -> x :: remove_dups l in
  fun l -> l |> List.sort compare |> remove_dups
